# TGE, the tarrasque, despite what DND would have you belive, is actually a turtle based creature.
import turtle
import random
import math

# CONSTANTS,
RADIANS_TO_DEGREES = 57.296
DEBUG_DRAWING = False
# NO MORE YELLING

		
class Transform():
	"""A class to encapsulate position, rotation and scale"""
	def __init__(self):
		self.position = Vector2(0,0)
		self.r = 0
		self.scale = 1.0
		
	# This moves the pen by the x,y coordinates, and rotates by r
	# Thus providing our worldspace offset
	def pen_to_pos(self):
		if DEBUG_DRAWING:
			turtle.pendown()
		turtle.forward(self.position.x)
		turtle.left(90)
		turtle.forward(self.position.y)
		turtle.left(-90)
		turtle.left(self.r)
		if DEBUG_DRAWING:
			turtle.penup()
	
	# Returns the turtle back from whence it came
	def return_pen_home(self):
		# Does not actually send pen home
		turtle.left(-self.r)
		turtle.left(90)
		turtle.backward(self.position.y)
		turtle.left(-90)
		turtle.backward(self.position.x)
		
class PhysicsBody():
	"""This was intended to grow into a basic collision body system, 
	but only holds movement data"""
	def __init__(self):
		self.radius = 1
		self.velocity = Vector2(0,0)
		self.ang_vel = 0

class Vector2():
	"""A two component vector, made by someone who can't do math"""
	def __init__(self, x, y) -> None:
		# intake and assure floatyness
		self.x = float(x)
		self.y = float(y)

	def __repr__(self) -> str:
		return f"({self.x}, {self.y})"

	def normalized(self):
		mag = self.magnitude()
		x = self.x / mag
		y = self.y / mag
		return Vector2(x,y)

	def magnitude(self):
		"""Returns the magnitude of the vector"""
		return math.sqrt(self.x ** 2 + self.y ** 2)

	def distance(self, vector2):
		"""Returns the distance between two point vectors."""
		return Vector2(self.x - vector2.x, self.y - vector2.y).magnitude()

	def dot_product(self, vector2):
		return self.x * vector2.x + self.y * vector2.y

	def angle(self, vector2):
		return math.acos(self.dot_product(vector2) / (self.magnitude() * vector2.magnitude()))
		
class GameObject():
	"""Super class from which active items inherit"""
	# Give us somewhere to be, and a way to get around
	def __init__(self):
		self.transform = Transform()
		self.body = PhysicsBody()
		
	# a default draw call to serve as an example and backup.
	def _draw(self):	
		# Pen is driven from 0,0 to our pos and rot
		self.transform.pen_to_pos()
		turtle.pendown()
		turtle.forward(10.0)
		turtle.left(90)
		turtle.forward(10)
		turtle.left(90)
		turtle.forward(20)
		turtle.left(90)
		turtle.forward(20)
		turtle.left(90)
		turtle.forward(20)
		turtle.left(90)
		turtle.forward(10)
		turtle.penup()
		turtle.left(-90)
		turtle.backward(10.0)
		# Pen is returned to home for next use.
		self.transform.return_pen_home()
				
	# A default _update method to facilitate going through the timestep
	# is usually called as a super method, for toroidal wrapping
	def _update(self): 
		# Move object by velocities 	
		# Physics sim
		self.transform.position.x += self.body.velocity.x * 0.033
		self.transform.position.y += self.body.velocity.y * 0.033
		self.transform.r += self.body.ang_vel * 0.33
		# Toroidal Planet Wrapping
#		if self.transform.x > 400:
#			self.transform.x -= 800
#		elif self.transform.x < -400:
#			self.transform.x += 800
#		if self.transform.y > 400:
#			self.transform.y -= 800
#		elif self.transform.y < -400:
#			self.transform.y += 800
	

class EngineInstance():
	def __init__(self) -> None:
		self.gameObjects = []
		self.frame_count = 0
		self.paused = True
		self.viewport_width = 1024
		self.viewport_height = 1024

		# Inputs, most not supported
		self.In_click_down = False
		self.In_click_held = False
		self.In_click_up = False
		self.In_lastClick = Vector2(0,0)
		self.In_mousePos = Vector2(0,0)

	def Start(self):
		self.paused = False
		# SECTION Turtle Setup
		turtle.setup(width= self.viewport_width, height=self.viewport_height)
		turtle.hideturtle()
		screen = turtle.Screen()
		screen.bgcolor('#000000')
		turtle.color("white")
		#turtle.speed(1) 
		turtle.tracer(False)
		# Basic Input
		turtle.getcanvas().bind('<Button 1>', self.mouseClick)
		turtle.getcanvas().bind('<Motion>',self.mouseMove)
		#turtle.getcanvas().bind('<esc>',self.pause)
		if self.paused == False:
			turtle.ontimer(self.gameloop, 33)
		# END SECTION Turtle Setup
		turtle.done()

	def mouseClick(self, event):
		self.In_lastClick = self.screenToWorldCoordinate(event.x, event.y)
		self.In_click_down = True
		#self.In_click_held = True # no support for held button currently
		# <Half A press joke here>

	def mouseMove(self, event):
		self.In_mousePos = self.screenToWorldCoordinate(event.x, event.y)

	def clearInputs(self):
		self.In_click_down = False

	def gameloop(self):
		turtle.clear()
		turtle.color(1,1,1)
		turtle.width(2)
		self.frame_count += 1
		# Reevaluate viewport
		self.viewport_width = turtle.window_width()
		self.viewport_height = turtle.window_height()
		# Update game objects
		for ob in self.gameObjects:
			ob._update()	
		# Draw them
		for ob in self.gameObjects:
			ob._draw()
		# Prepare next loop run
		if self.paused == False:
			turtle.ontimer(self.gameloop, 33)
		self.clearInputs()
	
	def pause(self):
		self.paused = not self.paused
		if self.paused == False:
			self.gameloop()
		
	def screenToWorldCoordinate(self, x, y):
		"""Transform a pixel screen coordinate to a pixel world coordinate"""
		x = x - self.viewport_width / 2
		y = -(y - self.viewport_height / 2)
		return Vector2(x, y)

	def addGameObject(self, new_GO):
		self.gameObjects.append(new_GO)

	def destroyGameObject(self, kill_GO):
		self.gameObjects.remove(kill_GO)
		# WARNING, LOTS OF GARBAGE

	

# SECTION Utility Functions
def clamp(value, min, max):
	if value < min:
		return min
	elif value > max:
		return max
	return value

# END SECTION Utility Functions


if __name__ == "__main__":
	import demo 
	demo.demo()
